﻿__copyright__ = "Copyright (c) 2014 CollabNet, Inc. All rights reserved."


import os
import logging
import urllib2
import time

from app.scripts import BaseInitializationScript
import app.util
import ext.lock

from suds.client import Client
from suds.cache import ObjectCache

class InitializationScript(BaseInitializationScript):

    scriptDescription = "Register SCM integration"

    def __getExternalSystemId(self, basePath):
        externalSystemId = None

        try:
            with open(os.path.join(basePath, '.scm.properties')) as scmProperties:
                for line in scmProperties:
                    key, val = line.partition("=")[::2]
                    if key == 'external_system_id':
                        externalSystemId = val.strip()
                        break
        except:
            pass
        return externalSystemId

    def __getService(self, soapBaseUrl, name):
        sudsCacheDir = os.path.join("%s" %self.cfg['TEMPORARY_DIR'], "suds")
        sudsCache = ObjectCache(sudsCacheDir)

        sudsLogFile = os.path.join(self.cfg['RUNTIME_LOG_DIR'], 'suds.log')
        logging.basicConfig(filename=sudsLogFile)
        logging.getLogger('suds.client').setLevel(self.log.getEffectiveLevel())

        wsdl = "%s/%s?wsdl" % (soapBaseUrl, name)
        return Client(wsdl, cache=sudsCache, location=wsdl)

    def __registerScmIntegration(self, soapBaseUrl, adapterType, name, title, paramNames, paramValues, externalSystemId):
        try:
            sessionId = self.__getService(soapBaseUrl, "CollabNet").service.login("scmadmin", self.cfg["SCM_ADMIN_PASSWORD"])
        except Exception as ex:
            raise Exception("Authenticating with TeamForge (at %s) failed: %r" % (soapBaseUrl, ex))

        scmService = self.__getService(soapBaseUrl, "ScmApp")
        adapterParams = scmService.factory.create('ns1:SoapNamedValues')
        adapterParams.names = paramNames
        adapterParams.values = paramValues

        if externalSystemId is not None:
            try:
                externalSystemSoapDO  = scmService.service.getExternalSystem(sessionId, externalSystemId )

                # Update the external system.
                self.log.info("Updating external system %s" % externalSystemId)
                if externalSystemSoapDO.systemDescription is None and (adapterType == "Subversion" or adapterType == "CVS"):
                    externalSystemSoapDO.systemDescription = title

                externalSystemSoapDO.adapterParams = adapterParams
                scmService.service.setExternalSystemData(sessionId, externalSystemSoapDO)

            except Exception as e:
                # Unfortunately, there is no error code we could rely on here.
                if "No entity found with id" in str(e):
                    # The system does not exist anymore.

                    raise ExternalSystemNotFoundException(
                        "Existing external system %s not found - either the settings in TeamForge have been tampered with " % externalSystemId +
                        "or the .scm.properties file in your SCM root is stale and should be removed")

                else:
                    # Different error, bail out.
                    raise e
        else:
            try:
                # Register new system.
                newExternalSystemId = scmService.service.addExternalSystem(
                    sessionId,
                    adapterType,
                    name,
                    title,
                    adapterParams)
                self.log.info("Registered %s with ExsyId %s" % (name, newExternalSystemId))
            except Exception as e:
                if "The parameter type/value 'RepositoryBaseUrl" in str(e):
                    # The system does not exist anymore.

                    raise ExternalSystemNotFoundException(
                        "Registration failed, most likely because the server's SSL certitificate is not trusted by CTF Core. " + str(e))
                
                else:
                    # Different error, bail out.
                    raise e
                

    def __isEndpointExternallyReachable(self, url):
        try:
            urllib2.urlopen(url, timeout=2).getcode()
            return True

        except urllib2.HTTPError, e:
            # Any code below 500 means that the server is reachable. A 5xx (503, in particular) could
            # come from a load balancer.
            return e.code < 500

        except Exception as e:
            self.log.warn("URL %s not externally reachable (yet): %r" % (url, e))
            return False

    def __waitForEndpointToBecomeExternallyReachable(self, url):
        # We know that the integration server is up - but if it is running behind a load balancer,
        # it may not be reachable by the application yet (the load balancer may take a few more seconds
        # before it takes the backend into the pool). However, if the integration server is not
        # reachable by the app, validation will fail and the registration will be rejected.
        #
        # Therefore, give the load balancer some slack.
        for i in range(1, 60):
            if self.__isEndpointExternallyReachable(url):
                return
            else:
                self.log.info("Endpoint '%s' not reachable yet, waiting..." % url)
                time.sleep(5)

        raise Exception("URL %s not externally reachable" % url)

    def __getCtfSoapUrl(self):
        return "%s://%s:%s/ce-soap60/services" % (
            self.cfg['HTTP_SCHEME'],
            self.cfg["integration/dependency/ctfcore-app/http-soap/remote-host"],
            self.cfg["integration/dependency/ctfcore-app/http-soap/port"])

    def __getIntegrationSoapUrl(self, serviceName):
        return "%s://%s:%s/" % (
            self.cfg['HTTP_SCHEME'],
            self.cfg["%s/cname/http-soap/host" % serviceName],
            self.cfg["%s/cname/http-soap/port" % serviceName])

    def __getIntegrationViewVcUrl(self, serviceName):
        scheme = self.cfg['HTTP_SCHEME']
        port = self.cfg["%s/cname/http-web-viewvc/port" % serviceName]

        if (scheme == "http" and port == "80") or (scheme == "https" and port == "443"):
            return "%s://%s/integration/viewvc/viewvc.cgi" % (
                scheme,
                self.cfg["%s/cname/http-web-viewvc/host" % serviceName])
        else:
            return "%s://%s:%s/integration/viewvc/viewvc.cgi" % (
                scheme,
                self.cfg["%s/cname/http-web-viewvc/host" % serviceName],
                port)

    def __getIntegrationHost(self, serviceName):
        return self.cfg["%s/cname/http-soap/host" % serviceName]

    def __getIntegrationPort(self, serviceName):
        return self.cfg["%s/cname/http-soap/port" % serviceName]

    def __getLocalHostName(self):
        return app.landscape.Landscape.getSite().getLocalMachine().getPhysicalName()

    def __registerSvn(self, svnRoot, externalSystemId):
        # Wait for CTF to be reachable for us (Note: By checking the URL, we ensure that not only Apache
        # has opened the port, but that Jboss is actually available.
        self.__waitForEndpointToBecomeExternallyReachable(self.__getCtfSoapUrl())

        # Wait for this server to be reachable by CTF.
        self.__waitForEndpointToBecomeExternallyReachable(self.__getIntegrationSoapUrl("subversion"))

        # Prepare parameter set. Note that we force the use of public FQDNs in order to avoid sidestepping
        # the load balancer in case # Integration and CTF Core happen to be installed on the same machine.
        params = {
            'ScmViewerUrl': self.__getIntegrationViewVcUrl("subversion"),
            'RepositoryRootPath': svnRoot,
            'RepositoryBaseUrl': "%s://%s%s" % (
                    self.cfg['HTTP_SCHEME'],
                    self.cfg["subversion/cname/http-svn/host"],
                    self.cfg['SUBVERSION_URI']),
            'HostSSL': "true" if self.cfg['SSL'] == 'on' else "false",
            'HostPort': self.__getIntegrationPort("subversion"),
            'HostName': self.__getIntegrationHost("subversion")
        }

        if externalSystemId is None:
            params['RequireApproval'] = 'false'

        paramNames = params.keys()
        paramValues = [params[p] for p in paramNames]

        # In case SVN is clustered, the two servers may be trying to register concurrently. This
        # can end up with two integrations being created. To avoid that, synchronize using a lock
        # on the (shared) SVN root.
        try:
            with ext.lock.NetworkFolderLock(svnRoot, self.__getLocalHostName()):
                # All right, we got the lock, so let us register.
                self.log.info("Registering SVN at %s with parameters %r" % (self.__getCtfSoapUrl(), params))

                self.__registerScmIntegration(
                    self.__getCtfSoapUrl(),
                    'Subversion',
                    "SVN",
                    "TeamForge Subversion SCM integration",
                    paramNames,
                    paramValues,
                    externalSystemId)

        except ext.lock.NetworkFolderOwnedByOtherPeer, e:
            # Lock acquisition failed.
            self.log.warn("The SVN root %s is locked by %s for registration - skipping registration" %
                (svnRoot, e.getPeer()))


    def __registerCvs(self, cvsRoot, externalSystemId):
        # Wait for CTF to be reachable for us (Note: By checking the URL, we ensure that not only Apache
        # has opened the port, but that Jboss is actually available.
        self.__waitForEndpointToBecomeExternallyReachable(self.__getCtfSoapUrl())

        # Wait for this server to be reachable by CTF.
        self.__waitForEndpointToBecomeExternallyReachable(self.__getIntegrationSoapUrl("cvs"))

        # Prepare parameter set. Note that we force the use of public FQDNs in order to avoid sidestepping
        # the load balancer in case # Integration and CTF Core happen to be installed on the same machine.

        params = {
            'ScmViewerUrl': self.__getIntegrationViewVcUrl("cvs"),
            'RepositoryRootPath': cvsRoot,
            'CvsHost': self.__getIntegrationHost("cvs"),
            'HostSSL': "true" if self.cfg['SSL'] == 'on' else "false",
            'HostPort': self.__getIntegrationPort("cvs"),
            'HostName': self.__getIntegrationHost("cvs")
        }

        if externalSystemId is None:
            params['RequireApproval'] = 'false'

        paramNames = params.keys()
        paramValues = [params[p] for p in paramNames]

        # In case CVS is clustered, the two servers may be trying to register concurrently. This
        # can end up with two integrations being created. To avoid that, synchronize using a lock
        # on the (shared) SVN root.
        try:
            with ext.lock.NetworkFolderLock(cvsRoot, self.__getLocalHostName()):
                # All right, we got the lock, so let us register.
                self.log.info("Registering CVS at %s with parameters %r" % (self.__getCtfSoapUrl(), params))

                self.__registerScmIntegration(
                    self.__getCtfSoapUrl(),
                    "CVS",
                    "CVS",
                    "TeamForge CVS SCM integration",
                    paramNames,
                    paramValues,
                    externalSystemId)


        except ext.lock.NetworkFolderOwnedByOtherPeer, e:
            # Lock acquisition failed.
            self.log.warn("The CVS root %s is locked by %s for registration - skipping registration" %
                (cvsRoot, e.getPeer()))


    def initialize(self):
        if app.landscape.Landscape.getSite().getLocalMachine().hasService('subversion') and \
                self.cfg.has_key('SUBVERSION_REPOSITORY_BASE') and \
                os.path.isdir(self.cfg['SUBVERSION_REPOSITORY_BASE']):

            self.__registerSvn(
                self.cfg['SUBVERSION_REPOSITORY_BASE'],
                self.__getExternalSystemId(self.cfg['SUBVERSION_REPOSITORY_BASE']))

        if app.landscape.Landscape.getSite().getLocalMachine().hasService('cvs') and \
                self.cfg.has_key('CVS_REPOSITORY_BASE') and \
                os.path.isdir(self.cfg['CVS_REPOSITORY_BASE']):

            self.__registerCvs(
                self.cfg['CVS_REPOSITORY_BASE'],
                self.__getExternalSystemId(self.cfg['CVS_REPOSITORY_BASE']))

class ExternalSystemNotFoundException(Exception):
    pass
